/*
  * Idea polega na utworzeniu ROZM_POCZ semaforow binarnych reprezentujacych
 * krzesla w poczekalni. Gdy klient wejdzie czeka na zwolnienie pierwszego
 * ale tylko w przypadku gdy semafor reprezentujacy ilosc wolnych krzesel
 * jest wiekszy od zera. Gdy klient juz usiadzie czeka na zwolnienie kolejnego
 * krzesla, gdy takowe sie zwolni, zwalnia on swoje poprzednio rezerwujac
 * kolejne. I tak az do fotelu fryzjera. Tworzona jest niejako kolejka
 * klientow zrealizowana na semaforach.
 * Gdy klient dotrze do fryzjera, budzi
 * go. Po skonczonym goleniu zwalnia ostatnie krzeslo i zwieksza ilosc wolnych
 * krzesel w poczekalni.
 * Fryzjer zas czeka nasygnal od klienta, ze dotarl na fotel.
 */

#define ROZM_POCZ 5
#define ILOSC_KLIENTOW 1000

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/stat.h>
#include "semafory.c"


int SemFryzjer; /* symbolizuje fryzjera */
int SemPojemnoscPoczekalni; /* ile jest wolnych krzesel */
int SemPoczekalnia; /* symbolizuje poczekalnie, ale nie jest to jeden
semafor tylko ROZM_POCZ semaforow binarnych, w ten sposob realizuje
kolejke procesow*/
int SemProcesy;  /* ile klientow jeszcze dziala */

static int NoWait0(int id) {
    struct sembuf bufor_sem;

    bufor_sem.sem_num = 0;
    bufor_sem.sem_op = -1;
    bufor_sem.sem_flg = SEM_UNDO | IPC_NOWAIT;

    if (semop(id, &bufor_sem, 1) == -1)
	return 1;

    return 0;
}


void klient(int numerKlienta) {
    int numer_w_kolejce = 0;

    /*najpierw spawdzam, czy jest miejsce wpoczekalni */
    if (NoWait0(SemPojemnoscPoczekalni)) {
        /* nie ma miejsca */
	printf("%d: Kurcze, pelna poczekalnia, spadam :nie wszedl\n", numerKlienta);
	return;
    }
    /* jest miejsce, moge wejsc */

    printf("%d: Wszedlem i czekam w kolejce.       :w kolejce\n", numerKlienta);

    /*pierwszy ruch w kolejce */
    if (WaitID(SemPoczekalnia, numer_w_kolejce++)){
	printf("%d: Blad 0!\n", numerKlienta);
	return;
    }

    /* a teraz poruszam sie dalej w kolejce */
    while(numer_w_kolejce < ROZM_POCZ) {
        /*najpierw czekam na wolne nastepne miejsce */
	if (WaitID(SemPoczekalnia, numer_w_kolejce)){
	    printf("%d: Blad 1!\n", numerKlienta);
	    return;
	}

        /*a potem zwalniam zajmowane */
	if (SignalID(SemPoczekalnia, (numer_w_kolejce++)-1)){
	    printf("%d: Blad 2!\n",numerKlienta);
	    return;
	}
    }

    /* koniec oczekiwania, jestem na fotelu.
     * Teraz trzeba obudzic fryzjera
     */
    printf("%d: Doszedlem do Fryzjera . Moja kolej!:bede obcinany\n", numerKlienta);
    Signal0(SemFryzjer);
    sleep(3); /* szybkie (3 s) golenie */

    printf("%d: Dziekuje, do widzenia!            :juz po obcieciu \n", numerKlienta);

    /*i zwalniam fotel */
    if(SignalID(SemPoczekalnia, numer_w_kolejce-1)){
	printf("%d: Blad 3!\n", numerKlienta);
	return;
    }

    /* oraz sygnalizuje, ze miejsce w poczekalni sie zwolnilo */
    Signal0(SemPojemnoscPoczekalni);

    return;
}

void Fryzjer(void) {
    while (1) {
	if(Wait0(SemFryzjer)){
	    /*czyli ktos ubil SemFryzjera*/
	    return;
	}
	printf("Fryzjer: Koniec drzemki, do dziela!  :wlasnie obcinam\n");
    }

}

int main(void) {
    int counter=0;
    int wywolujacy = 1;
    int nowyPid;

    SemPoczekalnia = utworzSemafor(ROZM_POCZ, 1);
    SemPojemnoscPoczekalni = utworzSemafor(1,ROZM_POCZ);
    SemFryzjer = utworzSemafor(1, 0);
    SemProcesy = utworzSemafor(1, 0);

    /* tworze ILOSC_KLIENTOW procesow klientow i 1 fryzjera */
    while(counter++ <= ILOSC_KLIENTOW && wywolujacy) {
	nowyPid = fork();

	switch(nowyPid) {
	case 0: /* jestesmy w procesie potomnym */
	    wywolujacy = 0;

	    if (counter == 1) /* pierwszy tworzony jest Fryzjer */
		Fryzjer();
	    else{  /* a reszta to zwykli, szarzy klienci */
                /* sygnalizuje utworzenie procesu */
		if (Signal0(SemProcesy)){
		    printf("c%d: Blad 1!\n", counter);
		    return 1;
		}

                /* uruchamiam klienta */
		klient(counter-1);

                /* sygnalizuje zakonczenie pracy procesu */
		if(Wait0(SemProcesy)){
		    printf("c%d: Blad2!\n", counter);
		    return 1;
		}
	    }

	    break;
	case -1:
	    printf("Blad podczas tworzenia podprocesow!\n");
	    return 1;
            break;
	default: /* jestesmy w procesie macierzystym */
	    wywolujacy = 1;

	    if (counter > ROZM_POCZ+1) /* gdy zrobilismy szybko kilka procesow nastepne beda tworzone troche szybciej */
		sleep(2);

	    break;
	}
    }

    if (wywolujacy) {

	if (WaitNull0(SemProcesy))
	    printf("Blad przy oczekiwanie na procesy klientow!\n");


	if (usunSemafor(SemProcesy) || usunSemafor(SemPoczekalnia) || usunSemafor(SemFryzjer) || usunSemafor(SemPojemnoscPoczekalni)){
	    printf("Blad podczas usuwania semaforow!\n");
	    return 1;
	}
    }

    /* szczesliwy koniec */
    return 0;
}
