/* Prosta implementacja interfejsu semaforow dla Unixa.
 
 */




/* Tworzy num prywatnych semaforow dla danego procesu
 * ktore potem inicjalizowane sa wartoscia val.
 * W przypadku sukcesu zwraca wartosc dodatnia, bedaca
 * identyfikatorem semafora;
 * w przeciwnym przypadku zwraca -1
 */
static int utworzSemafor(int num, int val);

/* Ustawia numer'ty semafor w zbiorze o identyfikatorze id wartoscia val.
 * Zwaraca 0 gdy wszystko przebieglo poprawnie, a 1 w przypadku bledu.
 */
static int ustawSemafor(int id, int val, int numer);

/* Usuwa cala kolekcje semaforow o identyfikatorze id.
 * Zwraca 0 w przypadku sukcesu, 1 w przypadku bledu.
 */
static int usunSemafor(int id);

/* Operacja Wait na zerowym semaforze w zbiorze o identyfikatorze id.
 * Zwraca 0 w przypadku sukcesu, 1 w przypadku bledu.
 */
static int Wait0(int id);

/* Jak wyzej, ale nie na zerowym tylko na num'tym semaforze.
 * Zwraca 0 w przypadku sukcesu, 1 w przypadku bledu.
 */
static int WaitID(int id, int num);

/* Operacja Wait nie zmieniajaca wartosci semafora, ale czekajaca
 * na semafor o wartosci 0.
 * Ta dziala na zerowym semaforze w kolekcji id.
 * Zwraca 0 w przypadku sukcesu, 1 w przypadku bledu.
 */
static int WaitNull0(int id);

/* Operacja Wait nie zmieniajaca wartosci semafora, ale czekajaca
 * na semafor o wartosci 0.
 * Ta dziala na semaforze o numerze num w kolekcji id.
 * Zwraca 0 w przypadku sukcesu, 1 w przypadku bledu.
 */
static int WaitNullID(int id, int num);

/* Operacja Signal na zerowym semaforze w zbiorze o identyfikatorze id.
 * Zwraca 0 w przypadku sukcesu, 1 w przypadku bledu.
 */
static int Signal0(int id);

/* Jak wyzej, ale nie na zerowym tylko na num'tym semaforze.
 * Zwraca 0 w przypadku sukcesu, 1 w przypadku bledu.
 */
static int SignalID(int id, int num);


/* Ponizsze linijki opcjonalnie definiuja unie semun
 * gdy nie jest zdefiniowana.
 * Ten fragment jest zywcem skopiowany z "man 2 semctl"
  */

#if defined(__GNU_LIBRARY__) && !defined(_SEM_SEMUN_UNDEFINED)
      /* union semun is defined by including <sys/sem.h> */
#else /* according to X/OPEN we have to define it ourselves */
union semun {
    int val;                    /* value for SETVAL */
    struct semid_ds *buf;    /* buffer for IPC_STAT, IPC_SET */
    unsigned short int *array;  /* array for GETALL, SETALL */
    struct seminfo *__buf;      /* buffer for IPC_INFO */
};
#endif


/* Tutaj znajduje sie implementacja funkcji opisanych na poczatku
 * tego pliku.
 */
static int ustawSemafor(int id, int val, int numer) {
    /* Pole val unii semun inicjujemy wartoscia val
     * ktore jest wykorzystywane przez komende SETVAL.
     */

    union semun unia_sem;

    unia_sem.val = val;

    if (semctl(id, numer, SETVAL, unia_sem) == -1)
	return 1;

    return 0;
}

static int usunSemafor(int id) {
    union semun unia_sem;

    if (semctl(id, 0, IPC_RMID, unia_sem) == -1)
	return 1;

    return 0;
}

static int Wait0(int id) {
    struct sembuf bufor_sem;

    bufor_sem.sem_num = 0; /* numer semafora w kolekcji */
    bufor_sem.sem_op = -1; /* operacja na semaforze, zgodnie ze spec. sem., -1 */
    bufor_sem.sem_flg = SEM_UNDO;

    if (semop(id, &bufor_sem, 1) == -1)
	return 1;

    return 0;
}

static int WaitID(int id, int num) {
    struct sembuf bufor_sem;

    bufor_sem.sem_num = num;
    bufor_sem.sem_op = -1;
    bufor_sem.sem_flg = SEM_UNDO;

    if (semop(id, &bufor_sem, 1) == -1)
	return 1;

    return 0;
}

static int WaitNull0(int id){
    struct sembuf bufor_sem;

    bufor_sem.sem_num = 0; /* numer semafora w kolekcji */
    bufor_sem.sem_op = 0; /* nic nie zmieniamy, tylko czekamy na sem. o wart. 0 */
    bufor_sem.sem_flg = SEM_UNDO;

    if (semop(id, &bufor_sem, 1) == -1)
	return 1;

    return 0;
}

static int WaitNullID(int id, int num){
    struct sembuf bufor_sem;

    bufor_sem.sem_num = num; /* numer semafora w kolekcji */
    bufor_sem.sem_op = 0; /* nic nie zmieniamy, tylko czekamy na sem. o wart. 0 */
    bufor_sem.sem_flg = SEM_UNDO;

    if (semop(id, &bufor_sem, 1) == -1)
	return 1;

    return 0;
}

static int Signal0(int id){
    struct sembuf bufor_sem;

    bufor_sem.sem_num = 0;
    bufor_sem.sem_op = 1; /* tutaj +1, zgodnie ze spec. semaforow */
    bufor_sem.sem_flg = SEM_UNDO;

    if (semop(id, &bufor_sem, 1) == -1)
	return 1;

    return 0;
}

static int SignalID(int id, int num){
    struct sembuf bufor_sem;

    bufor_sem.sem_num = num;
    bufor_sem.sem_op = 1;
    bufor_sem.sem_flg = SEM_UNDO;

    if (semop(id, &bufor_sem, 1) == -1)
	return 1;

    return 0;
}

static int utworzSemafor(int num, int val) {
    int sem_id;

    /* troche brzydko, ze na sztywno ustawiamy uprawnienia,
     * ale wystarcza to do podanego przykladu oraz do semaforow uzywanych
     * tylko przez potomstwo procesu macierzystego;
     * stad tez IPC_PRIVATE
     */
    sem_id = semget(IPC_PRIVATE, num, 0777);

    if (sem_id == -1)
	return  -1;

    /* gdy mamy juz semafor, trzeba zainicjalizowac kazdy z nowo utworzonych */
    while (--num >= 0){
	if (ustawSemafor(sem_id, val, num))
	    return -1;
    }

    return sem_id; /* gdy OK, zwracamy uzuskany identyfikator */
}
